/*
 * Copyright 1991 David A. Holland. All Rights Reserved.
 *
 * This software is provided 'as-is', without any express or implied
 * warranty.  In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 *
 *
 * (note from 2024: acknowledgement is definitely not worthwhile, don't bother)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "serial.h"
#include "xmodem.h"
//#include "lib.h"
#define log(level, ...) printf(__VA_ARGS__)
#define error(...) fprintf(stderr, __VA_ARGS__)


#define LOGGING 


#define STDTIME 10
#define HANDSHAKETIME 5
#define uchar unsigned char

#define SOH 1
#define EOT 4
#define ACK 6
#define NAK 21
#define CAN 24
#define BLOCKSIZE 128

static int crcflag; 
int crcpref = 1;

static unsigned docrc(unsigned crc, uchar val) {
  unsigned shifter,flag;
  if (crcflag) for (shifter = 0x80; shifter; shifter >>=1) {
    flag = (crc & 0x8000);
    crc <<= 1;
    crc |= ((shifter & val) ? 1 : 0);
    if (flag) crc ^= 0x1021;
  }
  else crc = (crc + val) & 255;
  return crc;
}

static void blocksend(uchar *bptr, unsigned blocknum, unsigned blocksize) {
  unsigned i, crc;
  #ifdef LOGGING
    log(DETAIL_LOG, "Sending block %d\n", blocknum);
  #endif
  qputc(SOH);
  qputc(blocknum);
  qputc(~blocknum);
  for (i=crc=0; i<blocksize; i++) {
    qputc(bptr[i]);
    crc = docrc(crc, bptr[i]);
  }
  crc = docrc(crc, 0);
  crc = docrc(crc, 0);
  if (crcflag) qputc((crc >> 8)&255);   /* checksum is only one byte */
  qputc(crc&255);
}

static int blockrcv(uchar *bptr, unsigned blocknum, unsigned blocksize) {
  unsigned i, crc;
  int bn, bad=0, store=0, x, c1, c2;
  #ifdef LOGGING
    log(DETAIL_LOG, "Receiving block %u\n", blocknum);
  #endif
  bn = qgetc(STDTIME);
  if (bn<0) return -1;
  #ifdef LOGGING
    log(DETAIL_LOG, "They're sending block %d\n", bn);
  #endif
  c1 = qgetc(STDTIME);
  if (c1<0) return -1;
  if (((~c1)&255) != (bn&255)) bad = 1;
  if ((bn&255)==(blocknum&255)) store = 1;
  for (i=crc=0; i<blocksize; i++) {
    x = qgetc(STDTIME);
    if (x<0) return -1;
    if (store) bptr[i] = x;
    crc = docrc(crc, bptr[i]);
  }
  c1 = qgetc(STDTIME);
  if (c1<0) return -1;
  if (crcflag) {
    crc = docrc(crc, 0);
    crc = docrc(crc, 0);
    c2 = qgetc(STDTIME);
    if (c2<0) return -1;
    #ifdef LOGGING
      log(DETAIL_LOG, "Our CRC: %04X   Theirs: %04X  bad=%d\n", crc, (c1<<8)|c2, bad);
    #endif
    if (crc!=(((unsigned)c1<<8)|(unsigned)c2)) bad = 3;
  }
  else {
    #ifdef LOGGING
      log(DETAIL_LOG, "Our checksum: %02X  Theirs: %02X  bad=%d\n", crc, c1, bad);
    #endif
    if ((unsigned)c1!=crc) bad = 3;
  }
  if (!store) bad=2;
  if (bad) error("Bad block, code %d (1=bad hdr 2=wrong block 3=crc err)\n", bad);
  return bad;
}

static int sendloop(uchar *bptr, unsigned nblocks, unsigned blocksize) {
  int ch, tries=0, naks=0;
  unsigned bn;
  crcflag = crcpref;
  #ifdef LOGGING
    log(DETAIL_LOG, "Handshaking (send)\n");
  #endif
  do {
    ch = qgetc(HANDSHAKETIME);
    #ifdef LOGGING
      log(DETAIL_LOG, "ch = %c [%d]; tries = %d\n", ch, ch, tries);
    #endif
    if (ch==CAN) return 1;
    tries++;
  }
  while (ch!=NAK && ch != 'C' && tries < 10);
  if (ch!=NAK && ch!='C') {error("Handshake failed\n"); return -1;}
  crcflag = (ch=='C');
  #ifdef LOGGING
    log(DETAIL_LOG, "Entering send loop\n");
  #endif
  for (bn=1; bn<=nblocks && naks<10;) {
    blocksend(bptr+(bn-1)*blocksize, bn, blocksize);
    ch = qgetc(STDTIME);
    if (ch<0) return -1;              /* loss of carrier, or timeout */
    else if (ch==CAN) return 1;       /* transfer aborted */
    else if (ch==ACK) {naks=0; bn++;} /* block received */
    else naks++;                      /* assume NAK */
  }
  #ifdef LOGGING
    log(DETAIL_LOG, "Transfer over\n");
  #endif
  do {
    qputc(EOT);
    ch = qgetc(STDTIME);
  }
  while (ch!=ACK && ch>=0);
  if (ch<0) return -1;
  return 0;
}

static uchar *rcvloop(unsigned blocksize, unsigned *bnum) {
  uchar *bptr=NULL;
  unsigned bn=1;
  int ch, tries=0, x;
  crcflag = crcpref;
  #ifdef LOGGING
    log(DETAIL_LOG, "Handshaking (rcv)\n");
  #endif
  do {
    if (tries==5) crcflag = !crcflag;
    qputc(crcflag ? 'C' : NAK);
    ch = qgetc(HANDSHAKETIME);
    tries++;
  }
  while (ch<0 && tries <10);
  if (ch<0) {error("Handshake failed\n"); return NULL;}
  #ifdef LOGGING
    log(DETAIL_LOG, "Entering receive loop\n");
  #endif
  while(1) {
    bptr = realloc(bptr, bn*blocksize);
    if (!bptr) {
      qputc(CAN);
      qputc(CAN);
      error("xmodem: Cannot allocate sufficient memory\n");
      return NULL;
    }
    if (ch==EOT) break;
    else if (ch==CAN) {error("xmodem: Aborted!\n"); return NULL;}
    else if (ch<0) {error("xmodem: Sector header timeout\n"); return NULL;}
    x = blockrcv(bptr+(bn-1)*blocksize, bn, blocksize);
    if (x==-1) return NULL;            /* timeout, loss of carrier */
    else if (x==1) qputc(NAK);         /* bad block */
    else if (x==2) qputc(ACK);         /* good block, wrong sector */
    else {qputc(ACK); bn++;}
    ch = qgetc(STDTIME); /* header for next block; we already had the first. */
  }
  #ifdef LOGGING
    log(DETAIL_LOG, "EOT received\n");
  #endif
  qputc(ACK);
  if (bnum) *bnum = bn;
  return bptr;
}

int xmodemsend(uchar *block) {
  return sendloop(block, 1, BLOCKSIZE);
}

int xmodemrcv(uchar *block) {
  uchar *foo;
  foo = rcvloop(BLOCKSIZE, NULL);
  if (foo) {
    memcpy(block, foo, BLOCKSIZE);
    free(foo);
    return 0;
  }
  else return 1;
}

int xmodemsend_f(char *filename) {
  FILE *f;
  long x, size;
  int v;
  uchar *p;
  log(MEDIUM_LOG, "Sending file %s\n", filename);
  f = fopen(filename, "rb");
  if (!f) {
    qputc(CAN);
    qputc(CAN);
    error("xmodem: Cannot open %s\n", filename);
    return 1;
  }
  fseek(f, 0, SEEK_END);
  x = ftell(f);
  size = (x + (BLOCKSIZE - (x % BLOCKSIZE)));
  p = malloc((unsigned)size);
  if (size > 65535L || !p) {
    qputc(CAN);
    qputc(CAN);
    error("xmodem: Cannot allocate memory for %s\n", filename);
    return 1;
  }
  memset(p, 0, (unsigned) size);
  fseek(f, 0, 0);
  fread(p, 1, (unsigned) size, f);
  fclose(f);
  v = sendloop(p, (unsigned) (size/BLOCKSIZE), BLOCKSIZE);
  free(p);
  return v;
}

int xmodemrcv_f(char *filename) {
  FILE *f;
  uchar *p;
  unsigned blocks;
  log(MEDIUM_LOG, "Receiving file %s\n", filename);
  f = fopen(filename, "wb");
  if (!f) {
    qputc(CAN);
    qputc(CAN);
    error("xmodem: Cannot open %s\n", filename);
    return 1;
  }
  p = rcvloop(BLOCKSIZE, &blocks);
  if (p) {
    fwrite(p, 1, blocks*BLOCKSIZE, f);
    free(p);
  }
  else error("Transfer failed\n");
  fclose(f);
  return (p==NULL);
}

/*
void usage(void) {
  printf("Usage:\n");
  printf("  xmd -rslcm filename\n");
  printf(" r = receive\n s = send\n l = log\n c = CRC\n m = checksum\n");
}

int main(int argc, char *argv[]) {
  int isreceive = 1, crcxfer = 1, i;
  if (argc==3 && *argv[1]=='-') {
    for (i=1; argv[1][i]; i++) switch(argv[1][i]) {
      case 'r' : isreceive = 1; break;
      case 's' : isreceive = 0; break;
      //case 'l' : log = 1; break;
      case 'c' : crcxfer = 1; break;
      case 'm' : crcxfer = 0; break;
      default: printf("Unknown switch %c\n", argv[1][i]); break;
    }
    crcflag = crcxfer;
    qinit();
    if (isreceive) xmodemrcv_f(argv[2]);
    else xmodemsend_f(argv[2]);
    qclose();
  }
  else usage();
  return 0;
}
*/
